/* derived from jkeztrans.c and jktrans.c */

#include <windows.h>
#include <process.h>
#include <stdio.h>
#include <string.h>

#define OTHEXPORT __declspec(dllexport)

#include "oh_trmod.h"

/* keep BUFFSIZE small because of translation software bug. */
#define BUFFSIZE 800
#define OTHPATHLEN 2048

#define OTHLOCKINIT() InitializeCriticalSection(&CriticalSection)
#define OTHLOCK() EnterCriticalSection(&CriticalSection)
#define OTHUNLOCK() LeaveCriticalSection(&CriticalSection)
#define OTHLOCKDEL() DeleteCriticalSection(&CriticalSection)

static LPWSTR uselesschars[] =
  {
    (LPWSTR)"\x09\x00",
    (LPWSTR)"\x0a\x00",
    (LPWSTR)"\x0d\x00",
    (LPWSTR)"\x20\x00",
    (LPWSTR)"\x00\x30"
  };
static const int nuselesschars = sizeof(uselesschars) / sizeof(LPWSTR);

static unsigned int iJKTransServerAddress;
static unsigned int iJKTransServerPort = 80;
static char *lpszJKTransServerName = "ocn.amikai.com";
static char lpszJKTransQuery[10000] =
  "POST /amitext/indexUTF8.jsp HTTP/1.0\n"
  "Host: ocn.amikai.com\n"
  "Content-Type: application/x-www-form-urlencoded\n"
  "Content-Length:      \n\n"
  "c_id=ocn&lang=EN&langpair=JA%2CEN&translatedText=&translate=Translate"
  "&sourceText=";

static HINSTANCE hInst;
static HMODULE hBabel;
static int isRunning;

static CRITICAL_SECTION CriticalSection;
static WCHAR jpbuff[BUFFSIZE];
static int jpendpos;
static WCHAR krbuff[BUFFSIZE];
static int krendpos;
static HANDLE jktrth;

static void OthStrStr(LPWSTR s2, LPWSTR *s3);
static LPWSTR OthJKTrans(LPWSTR jpStr);
static void OthPushJP(LPCWSTR jpStr, int len);
static LPWSTR OthPullJP(void);
static void OthPushKR(LPWSTR krStr);
static LPWSTR OthPullKR(void);

static void OthJKTransThread(void *arg);

/* hack to optimize */
static void OthStrStr(LPWSTR s2, LPWSTR *s3)
{
  LPWSTR s1;
  for (s1 = jpbuff + jpendpos; s1 != jpbuff && s1 > *s3; --s1)
  {
    if (*s1 == *s2)
    {
      *s3 = s1;
    }
  }
}

static LPWSTR OthJKTrans(LPWSTR jpStr)
{
  struct sockaddr_in sock;
  int sd;
  char *recvbuf;
  int recvbufsize;
  int i, j, len;
  int timeout = 10000;
  char utfjp[BUFFSIZE * 4];
  char *s1;
  char *s_token = "<textarea name=\"translatedText\""
    " class=\"translateTextarea\" rows=\"15\" cols=\"30\" wrap=\"virtual\">";
  char *e_token = "</textarea>";
  LPWSTR retstr;

  len = lstrlenW(jpStr);
  for (j = 0; j < nuselesschars; ++j)
  {
    for (i = 0; i < len && jpStr[0] != 0; ++i)
    {
      if (jpStr[i] == uselesschars[j][0])
      {
        memmove(&jpStr[i], &jpStr[i + 1], (len - i) << 1);
        --len;
        --i;
      }
    }
  }

  if (jpStr[0] == 0)
  {
    return NULL;
  }

  WideCharToMultiByte(CP_UTF8, 0, jpStr, -1, utfjp, BUFFSIZE * 4, 0, 0);

  s1 = strstr(lpszJKTransQuery, "&sourceText=");
  s1 = s1 + 12;
  for (i = 0, j = 0; utfjp[i] != '\0'; ++i)
  {
    sprintf(&s1[j], "%%%02x", (unsigned char)utfjp[i]);
    j += 3;
  }
  s1[j] = '\x0a';
  ++j;
  s1[j] = '\0';
  s1 = strstr(lpszJKTransQuery, "c_id=ocn&lang=EN");
  j = strlen(s1);
  s1 = strstr(lpszJKTransQuery, "Content-Length:");
  s1 = s1 + 16;
  for (i = 0; i < 5; ++i)
  {
    s1[i] = ' ';
  }
  sprintf(s1, "%d", j);
  for (i = 0; i < 5; ++i)
  {
    if (s1[i] == '\0')
    {
      s1[i] = ' ';
    }
  }

  memset(&sock, 0, sizeof(sock));
  sock.sin_family = AF_INET;
  sock.sin_addr.s_addr = iJKTransServerAddress;
  sock.sin_port = htons((unsigned short)iJKTransServerPort);

  if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
  {
    return NULL;
  }

  if (setsockopt(sd, SOL_SOCKET, SO_RCVTIMEO, (const char *)&timeout,
    sizeof(timeout)) == SOCKET_ERROR)
  {
    closesocket(sd);
    return NULL;
  }

  if (connect(sd, (struct sockaddr *)&sock, sizeof(sock)) == -1)
  {
    closesocket(sd);
    return NULL;
  }

  if (send(sd, lpszJKTransQuery, strlen(lpszJKTransQuery), 0) == -1)
  {
    closesocket(sd);
    return NULL;
  }

  recvbuf = malloc(BUFFSIZE * 4);
  recvbufsize = BUFFSIZE * 4;

  i = 0;

  while ((j = recv(sd, &recvbuf[i], BUFFSIZE * 4, 0)) > 0)
  {
    i += j;
    if (i + BUFFSIZE * 4 > recvbufsize)
    {
      recvbuf = realloc(recvbuf, (recvbufsize = recvbufsize + BUFFSIZE * 4));
      if (recvbuf == NULL)
      {
        closesocket(sd);
        return NULL;
      }
    }
  }

  closesocket(sd);

  i = strlen(s_token);
  s1 = strstr(recvbuf, s_token);
  s1 = s1 + i;
  *strstr(s1, e_token) = '\0';
  i = strlen(s1);
  retstr = malloc(i * 4);
  MultiByteToWideChar(CP_UTF8, 0, s1, -1, retstr, i * 2);
  free(recvbuf);

  len = lstrlenW(retstr);
  for (i = 0; i < len; ++i)
  {
    if (retstr[i] == 0x000a)
    {
      memmove(&retstr[i + 1], &retstr[i], (len - i + 1) << 1);
      retstr[i] = 0x000d;
      ++len;
      ++i;
    }
  }

  return retstr;
}

static void OthPushJP(LPCWSTR jpStr, int len)
{
  int i;

  if (len <= 0)
    return;

  OTHLOCK();

  if (jpendpos >= BUFFSIZE)
  {
    jpbuff[0] = 0;
    jpendpos = 0;
  }

  for (i = 0; i < len && jpendpos < BUFFSIZE && jpStr[i] != 0;
    ++i, ++jpendpos)
  {
    jpbuff[jpendpos] = jpStr[i];
  }

  if (jpendpos >= BUFFSIZE)
  {
    jpbuff[0] = 0;
    jpendpos = 0;
  }

  jpbuff[jpendpos] = 0;

  OTHUNLOCK();
}

static LPWSTR delimiters[] =
  {
    (LPWSTR)"\x21\x00",
    (LPWSTR)"\x22\x00",
    (LPWSTR)"\x27\x00",
    (LPWSTR)"\x28\x00",
    (LPWSTR)"\x29\x00",
    (LPWSTR)"\x2c\x00",
    (LPWSTR)"\x2e\x00",
    (LPWSTR)"\x3a\x00",
    (LPWSTR)"\x3b\x00",
    (LPWSTR)"\x3c\x00",
    (LPWSTR)"\x3e\x00",
    (LPWSTR)"\x3f\x00",
    (LPWSTR)"\x5b\x00",
    (LPWSTR)"\x5d\x00",
    (LPWSTR)"\x5e\x00",
    (LPWSTR)"\x60\x00",
    (LPWSTR)"\x7b\x00",
    (LPWSTR)"\x7d\x00",
    (LPWSTR)"\xab\x00",
    (LPWSTR)"\xbb\x00",
    (LPWSTR)"\xb4\x00",
    (LPWSTR)"\xb7\x00",
    (LPWSTR)"\xb8\x00",
    (LPWSTR)"\x01\x30",
    (LPWSTR)"\x02\x30",
    (LPWSTR)"\x0c\xff",
    (LPWSTR)"\x0e\xff",
    (LPWSTR)"\xfb\x30",
    (LPWSTR)"\x1a\xff",
    (LPWSTR)"\x1b\xff",
    (LPWSTR)"\x1f\xff",
    (LPWSTR)"\x01\xff",
    (LPWSTR)"\x9b\x30",
    (LPWSTR)"\x9c\x30",
    (LPWSTR)"\xb4\x00",
    (LPWSTR)"\x40\xff",
    (LPWSTR)"\xa8\x00",
    (LPWSTR)"\x3e\xff",
    (LPWSTR)"\x18\x20",
    (LPWSTR)"\x19\x20",
    (LPWSTR)"\x1c\x20",
    (LPWSTR)"\x1d\x20",
    (LPWSTR)"\x08\xff",
    (LPWSTR)"\x09\xff",
    (LPWSTR)"\x14\x30",
    (LPWSTR)"\x15\x30",
    (LPWSTR)"\x3b\xff",
    (LPWSTR)"\x3d\xff",
    (LPWSTR)"\x5b\xff",
    (LPWSTR)"\x5d\xff",
    (LPWSTR)"\x08\x30",
    (LPWSTR)"\x09\x30",
    (LPWSTR)"\x0a\x30",
    (LPWSTR)"\x0b\x30",
    (LPWSTR)"\x0c\x30",
    (LPWSTR)"\x0d\x30",
    (LPWSTR)"\x0e\x30",
    (LPWSTR)"\x0f\x30",
    (LPWSTR)"\x10\x30",
    (LPWSTR)"\x11\x30",
    (LPWSTR)"\x1c\xff",
    (LPWSTR)"\x1e\xff",
    (LPWSTR)"\x32\x20",
    (LPWSTR)"\x33\x20",
    (LPWSTR)"\x6a\x22",
    (LPWSTR)"\x6b\x22",
    (LPWSTR)"\x26\x20",
    (LPWSTR)"\x1a\x20",
    (LPWSTR)"\x1e\x20",
    (LPWSTR)"\xc6\x02",
    (LPWSTR)"\x39\x20",
    (LPWSTR)"\x18\x20",
    (LPWSTR)"\x19\x20",
    (LPWSTR)"\x1c\x20",
    (LPWSTR)"\x1d\x20",
    (LPWSTR)"\x22\x20",
    (LPWSTR)"\x3a\x20",
    (LPWSTR)"\x62\xff",
    (LPWSTR)"\x63\xff",
    (LPWSTR)"\x64\xff"
  };
static const int ndelimiters = sizeof(delimiters) / sizeof(LPWSTR);

static LPWSTR OthPullJP(void)
{
  int dpos, i;
  LPWSTR newdposptr = NULL;
  LPWSTR retjp = NULL;

  OTHLOCK();

  if (jpendpos == 0)
  {
    goto ret;
  }

  for (i = 0; i < ndelimiters; i++)
  {
    OthStrStr(delimiters[i], &newdposptr);
  }

  if (newdposptr == NULL || newdposptr == jpbuff)
  {
    retjp = NULL;
  }
  else
  {
    dpos = newdposptr - jpbuff + 1;
    retjp = malloc((dpos + 3) << 1);
    retjp[0] = 0x000d;
    retjp[1] = 0x000a;
    memcpy(retjp + 2, jpbuff, dpos << 1);
    retjp[dpos + 2] = 0;

    jpendpos = lstrlenW(jpbuff + dpos);
    memmove(jpbuff, jpbuff + dpos, (jpendpos + 1) << 1);
  }

ret:

  OTHUNLOCK();

  return retjp;
}

static void OthPushKR(LPWSTR krStr)
{
  int i, len;

  if (krStr == NULL)
    return;

  len = lstrlenW(krStr);

  OTHLOCK();

  if (krendpos >= BUFFSIZE)
  {
    krbuff[0] = 0;
    krendpos = 0;
  }

  for (i = 0; i < len && krendpos < BUFFSIZE; ++i, ++krendpos)
  {
    krbuff[krendpos] = krStr[i];
  }

  if (krendpos >= BUFFSIZE)
  {
    krbuff[0] = 0;
    krendpos = 0;
  }

  krbuff[krendpos] = 0;

  OTHUNLOCK();
}

static LPWSTR OthPullKR(void)
{
  LPWSTR ret = NULL;
  int len;

  OTHLOCK();

  if (krendpos != 0)
  {
    len = lstrlenW(krbuff);
    ret = malloc((len + 1) * 2);
    lstrcpyW(ret, krbuff);
    krbuff[0] = 0;
    krendpos = 0;
  }

  OTHUNLOCK();

  return ret;
}

static void OthJKTransThread(void *arg)
{
  LPWSTR jpStr;
  LPWSTR krStr;

  while (isRunning)
  {
    jpStr = OthPullJP();
    if (jpStr != NULL)
    {
      if (jpStr[0] != 0)
      {
        krStr = OthJKTrans(jpStr);
        OthPushKR(krStr);
        free(krStr);
      }
      free(jpStr);
    }
    Sleep(1000);
  }
}

OTHEXPORT int OTHCALL OthTransInit(OTHTR *pTR)
{
  struct hostent *hp;
  WSADATA wsaData;

  if (isRunning)
  {
    ++isRunning;
    return 0;
  }

  pTR->lpStr = NULL;
  pTR->iLen = 0;

  WSAStartup(MAKEWORD(2, 0), &wsaData);

  hp = gethostbyname(lpszJKTransServerName);

  if (hp == 0)
    return -1;

  iJKTransServerAddress = ((struct in_addr *)(hp->h_addr))->s_addr;

  OTHLOCKINIT();

  isRunning = 1;

  jktrth = (HANDLE)_beginthread(&OthJKTransThread, 0, NULL);

  return 0;
}

OTHEXPORT int OTHCALL OthTransSetting(OTHTR *pTR)
{
  return 0;
}

OTHEXPORT int OTHCALL OthTransCopySetting(OTHTR *pDestTR, OTHTR *pSrcTR)
{
  return 0;
}

OTHEXPORT int OTHCALL OthTranslate(OTHTR *pTR, LPCWSTR lpStr, int iLen)
{
  LPWSTR krStr;

  if (lpStr[0] != 0)
  {
    OthPushJP(lpStr, iLen);
  }

  krStr = OthPullKR();

  free(pTR->lpStr);
  pTR->lpStr = NULL;

  if (krStr)
  {
    pTR->lpStr = krStr;
    pTR->iLen = lstrlenW(krStr);
  }

  if (pTR->lpStr == NULL)
  {
    pTR->lpStr = malloc(2);
    pTR->lpStr[0] = 0;
    pTR->iLen = 0;
  }

  return 0;
}

OTHEXPORT int OTHCALL OthTransDestroy(OTHTR *pTR)
{
  --isRunning;
  if (isRunning != 0)
    return 0;

  if (jktrth)
  {
    WaitForSingleObject(jktrth, INFINITE);
    jktrth = NULL;
  }
  jpbuff[0] = '\0';
  jpendpos = 0;
  krbuff[0] = '\0';
  krendpos = 0;
  OTHLOCKDEL();
  free(pTR->lpStr);

  return 0;
}

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
    hInst = hinstDLL;

  return TRUE;
}
